// -------------------------------------------------------------------------------- 
//!  @file  common.h                                                                
//!  共通処理・データのヘッダファイル                                               
//                                                                                  
//   Copyright (C) 2022 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#ifndef __COMMON_H__INCLUDED__
#define __COMMON_H__INCLUDED__


#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <pico/stdlib.h>
#include "config.h"
#include "libvga/vga/vga.h"
#include "libvga/pad/pad.h"
#include "libvga/dac/mcp49x1.h"
#include "libvga/pcm/sn76489.h"

#define _THISPTR(_n, _m, _p)		((_n *)((uint8_t *)(_p) - (uint8_t *)&((_n *)0)->_m))

#define __enable_irq()				__asm volatile ("cpsie i")		//!< 割り込み許可
#define __disable_irq()				__asm volatile ("cpsid i")		//!< 割り込み禁止

#define writePin(p, value)					(p)->write((p), value)
#define readAddressBus(p, address)			(p)->read((p), address)
#define writeAddressBus(p, address, data)	(p)->write((p), address, data)


typedef struct _pin					PIN ;
typedef struct _addressbus			ADDRESSBUS ;
typedef struct _clockevent			CLOCKEVENT ;

typedef int32_t (*CLOCKEVENTPROC)( CLOCKEVENT *p_this ) ;			//!< クロックイベント処理


	enum {
		IMAGE_WIDTH					=	256,						//!< 画像幅
		IMAGE_HEIGHT				=	192,						//!< 画像高さ
		SM3_WIDTH					=	256,						//!< SEGA MARKIII 表示幅
		SM3_HEIGHT					=	192,						//!< SEGA MARKIII 表示高さ
		GG_WIDTH					=	160,						//!< GAME GEAR 表示幅
		GG_HEIGHT					=	144,						//!< GAME GEAR 表示高さ
		GG_MARGIN_TOP				=	24,							//!< GAME GEAR 上余白
		GG_MARGIN_BOTTOM			=	24,							//!< GAME GEAR 下余白
		GG_MARGIN_LEFT				=	48,							//!< GAME GEAR 左余白
		GG_MARGIN_RIGHT				=	48,							//!< GAME GEAR 右余白
	} ;

	//! 動作モード
	enum {
		MODE_SM3,													//!< SEGA MARK III
		MODE_GG,													//!< GAME GEAR
		MODE_MAX
	} ;

	//! パッド種類
	enum {
		PAD_TYPE_NORMAL,											//!< 通常
		PAD_TYPE_PADDLE												//!< パドルコントロール
	} ;

	//! VRAM
	typedef struct {
		uint16_t					palette[IMAGE_HEIGHT][32] ;		//!< パレット
		uint8_t						image[IMAGE_HEIGHT][IMAGE_WIDTH] ;
																	//!< 画像
		uint16_t					borderColor ;					//!< 画面上下ボーダー色
		uint16_t					dummy ;
	} VRAM ;

	//! ラピッドファイア
	typedef struct {
		uint8_t						button ;						//! ボタン番号
		uint8_t						count ;							//! カウンタ
	} RAPIDFIRE ;

	//! ピンインタフェース
	struct _pin {
		void (*write)(PIN *p_this, int16_t value) ;					//!< 書き込み
	} ;

	//! アドレスバスインタフェース
	struct _addressbus {
		uint8_t (*read)(ADDRESSBUS *p_this, uint16_t address) ;		//!< 読み込み
		void    (*write)(ADDRESSBUS *p_this, uint16_t address, uint8_t data) ;
																	//!< 書き込み
	} ;

	//! クロックイベントインタフェース
	struct _clockevent {
		CLOCKEVENT					*p_next ;						//!< 次のクロックイベント
		uint32_t					clock ;							//!< 残りクロック数
		int32_t (*process)( CLOCKEVENT *p_this ) ;					//!< 処理
	} ;

	//! ROM 情報
	typedef struct {
		uint8_t						const *p_rom ;					//!< ROM
		uint32_t					romSize ;						//!< ROM サイズ
		uint8_t						ramSize ;						//!< RAM サイズ上位１バイト
		uint8_t						mode ;							//!< 動作モード
		char						name[22] ;						//!< 表示名
	} ROMINFO ;


#include "rominfo.h"
#include "z80.h"
#include "vdp.h"
#include "sound.h"
#include "sm3gg.h"

	//! グローバル変数
	typedef struct {
		uint8_t						menu ;							//!< メニュー表示フラグ
		uint8_t						proc ;							//!< メニュー処理番号
		int8_t						select ;						//!< ゲーム選択番号
		int8_t						cursor ;						//!< カーソル番号
		uint8_t						wide ;							//!< 画面ワイドフラグ
		uint8_t						language ;						//!< 言語フラグ
		uint8_t						game ;							//!< ゲーム番号
		uint8_t						mode ;							//!< 動作モード
		uint8_t						page ;							//!< VRAM カレントページ
		uint8_t						borderColor ;					//!< ボーダー色
		RAPIDFIRE					rapidFire[2] ;					//!< ラピッドファイア

		//! パッド状態
		struct {
			uint8_t					type ;							//!< パッド種類
			uint8_t					analog ;						//!< アナログ値
			uint16_t				norm ;							//!< 通常
			uint16_t				old ;							//!< 以前
			uint16_t				once ;							//!< １度のみ

			//! パドルコントロール
			struct {
				uint8_t				pdsel ;							//!< PDSEL カウンタ
				uint8_t				pos ;							//!< 位置
			} paddle ;
		} pad ;

		VRAM						*p_vram ;						//!< 表示 VRAM
		VRAM						vram[2] ;						//!< VRAM
		uint16_t					txt[VGA_TXT_HEIGHT_COUNT][VGA_TXT_WIDTH_COUNT] ;
																	//!< テキスト VRAM
		CLOCKEVENT					*p_clockEventList ;				//!< クロックイベントリスト
		uint32_t					baseClock ;						//!< ベースクロック
	} GLOBALVAR ;


extern GLOBALVAR					g ;
extern uint16_t						g_fontPalette[] ;
extern uint32_t						g_fontPattern[][8] ;


extern void         common_init( void ) ;
extern void RAMFUNC clockgen_attachEvent(CLOCKEVENT *p_clockEvent, uint32_t clock) ;
extern void RAMFUNC clockgen_supplyClock( void ) ;


#endif//__COMMON_H__INCLUDED__
