// -------------------------------------------------------------------------------- 
//!  @file  vdp.c                                                                   
//!  VDP のソースファイル                                                           
//                                                                                  
//   Copyright (C) 2022 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"

//! 行データ
#define LINEDATA(func, clk)			(((clk)*2)<<4)|(LINEPROC_ ## func)

	//! VDP 情報
	enum {
		VDP_RBLANKING				=	8,							//!< 右ブランキング期間
		VDP_HSYNC					=	26,							//!< HSync 期間
		VDP_LBLANKING				=	24,							//!< 左ブランキング期間
		VDP_LBORDER					=	13,							//!< 左ボーダー期間
		VDP_HACTIVE					=	256,						//!< 水平アクティブディスプレイ期間
		VDP_RBORDER					=	15,							//!< 右ボーダー期間
		VDP_BBLANKING				=	3,							//!< 下ブランキング期間
		VDP_VSYNC					=	3,							//!< VSync 期間
		VDP_TBLANKING				=	13,							//!< 上ブランキング期間
		VDP_TBORDER					=	27,							//!< 上ボーダー期間
		VDP_VACTIVE					=	192,						//!< 垂直アクティブディスプレイ期間
		VDP_BBORDER					=	24,							//!< 下ボーダー期間
		VDP_LMARGIN					=	VDP_RBLANKING+VDP_HSYNC+VDP_LBLANKING,
																	//!< 左マージン
		VDP_TMARGIN					=	VDP_BBLANKING+VDP_VSYNC+VDP_TBLANKING,
																	//!< 上マージン
		VDP_VSCREEN_WIDTH			=	342,						//!< 仮想画面横幅
		VDP_VSCREEN_HEIGHT			=	262,						//!< 仮想画面縦幅
																	//!< 画面横幅
		VDP_SCREEN_WIDTH			=	VDP_LBORDER+VDP_HACTIVE+VDP_RBORDER,
																	//!< 画面縦幅
		VDP_SCREEN_HEIGHT			=	VDP_TBORDER+VDP_VACTIVE+VDP_BBORDER,
		VDP_VIEW_WIDTH				=	512,						//!< VRAM 表示横幅
		VDP_VIEW_HEIGHT				=	256,						//!< VRAM 表示縦幅
		VDP_VRAM_SIZE				=	0x4000,						//!< VRAM サイズ
		VDP_SPRITE_MAX				=	64							//!< スプライト最大数
	} ;

	//! コマンド
	enum {
		CMD_READVRAM				=	0,							//!< VRAM 読み取り
		CMD_WRITEREG				=	2,							//!< レジスタ書き込み
		CMD_WRITECRAM				=	3,							//!< CRAM 書き込み
	} ;

	//! ステータス
	enum {
		STATUS_F_					=	7,							//!< 割り込みフラグ
		STATUS_5S_					=	6,							//!< スプライト数オーバー
		STATUS_C_					=	5,							//!< スプライト衝突フラグ
		STATUS_F					=	1 << STATUS_F_,
		STATUS_5S					=	1 << STATUS_5S_,
		STATUS_C					=	1 << STATUS_C_
	} ;

	//! レジスタ０
	enum {
		R0_VSI_						=	7,							//!< 画面右 64dot 固定
		R0_HSI_						=	6,							//!< 画面上 16dot 固定
		R0_LCB_						=	5,							//!< 画面左  8dot マスク
		R0_IE1_						=	4,							//!< HSync 割り込み有効
		R0_SS_						=	3,							//!< スプライト左 8dot シフト
		R0_M4_						=	2,							//!< モード選択ビット４
		R0_M3_						=	1,							//!< モード選択ビット３
		R0_EV_						=	0,							//!< 外部ビデオ信号
		R0_VSI						=	1 << R0_VSI_,
		R0_HSI						=	1 << R0_HSI_,
		R0_LCB						=	1 << R0_LCB_,
		R0_IE1						=	1 << R0_IE1_,
		R0_SS						=	1 << R0_SS_,
		R0_M4						=	1 << R0_M4_,
		R0_M3						=	1 << R0_M3_,
		R0_EV						=	1 << R0_EV_
	} ;

	//! レジスタ１
	enum {
		R1_4_16K_					=	7,							//!< 4/16K VRAM 選択
		R1_BLANK_					=	6,							//!< 画面表示
		R1_IE_						=	5,							//!< INT 有効
		R1_M1_						=	4,							//!< モード選択ビット１
		R1_M2_						=	3,							//!< モード選択ビット２
		R1_SIZE_					=	1,							//!< スプライトサイズ
		R1_MAG_						=	0,							//!< スプライト拡大
		R1_4_16K					=	1 << R1_4_16K_,
		R1_BLANK					=	1 << R1_BLANK_,
		R1_IE						=	1 << R1_IE_,
		R1_M1						=	1 << R1_M1_,
		R1_M2						=	1 << R1_M2_,
		R1_SIZE						=	1 << R1_SIZE_,
		R1_MAG						=	1 << R1_MAG_
	} ;

	//! 行処理
	enum {
		LINEPROC_RBLANK_BEGIN,
		LINEPROC_RBLANK,
		LINEPROC_RBLANK_DRAW,
		LINEPROC_RBLANK_DRAW_END,
		LINEPROC_VCOUNT,
		LINEPROC_VSYNC,
		LINEPROC_VSYNC_FLAG,
		LINEPROC_HSYNC,
		LINEPROC_YPOS,
		LINEPROC_OVR_FLAG
	} ;


static uint32_t						m_planeToPixel[2][256] ;		//! プレーンピクセル変換テーブル

//! H カウンタ
static uint8_t						m_hcount[VDP_VSCREEN_WIDTH]={
										// RBLANKING
										0x87,0x88,0x88,0x89,0x89,0x8a,0x8a,0x8b,

										// HSYNC
										0x8b,0x8c,0x8c,0x8d,0x8d,0x8e,0x8e,0x8f,0x8f,0x90,0x90,0x91,0x91,0x92,0x92,0x93,
										0x93,0xe9,0xe9,0xea,0xea,0xeb,0xeb,0xec,0xec,0xed,

										// LBLANKING
										0xed,0xee,0xee,0xef,0xef,0xf0,0xf0,0xf1,0xf1,0xf2,0xf2,0xf3,0xf3,0xf4,0xf4,0xf5,
										0xf5,0xf6,0xf6,0xf7,0xf7,0xf8,0xf8,0xf9,

										// LBORDER
										0xf9,0xfa,0xfa,0xfb,0xfb,0xfc,0xfc,0xfd,0xfd,0xfe,0xfe,0xff,0xff,

										// HACTIVE
										0x00,0x00,0x01,0x01,0x02,0x02,0x03,0x03,0x04,0x04,0x05,0x05,0x06,0x06,0x07,0x07,
										0x08,0x08,0x09,0x09,0x0a,0x0a,0x0b,0x0b,0x0c,0x0c,0x0d,0x0d,0x0e,0x0e,0x0f,0x0f,
										0x10,0x10,0x11,0x11,0x12,0x12,0x13,0x13,0x14,0x14,0x15,0x15,0x16,0x16,0x17,0x17,
										0x18,0x18,0x19,0x19,0x1a,0x1a,0x1b,0x1b,0x1c,0x1c,0x1d,0x1d,0x1e,0x1e,0x1f,0x1f,
										0x20,0x20,0x21,0x21,0x22,0x22,0x23,0x23,0x24,0x24,0x25,0x25,0x26,0x26,0x27,0x27,
										0x28,0x28,0x29,0x29,0x2a,0x2a,0x2b,0x2b,0x2c,0x2c,0x2d,0x2d,0x2e,0x2e,0x2f,0x2f,
										0x30,0x30,0x31,0x31,0x32,0x32,0x33,0x33,0x34,0x34,0x35,0x35,0x36,0x36,0x37,0x37,
										0x38,0x38,0x39,0x39,0x3a,0x3a,0x3b,0x3b,0x3c,0x3c,0x3d,0x3d,0x3e,0x3e,0x3f,0x3f,
										0x40,0x40,0x41,0x41,0x42,0x42,0x43,0x43,0x44,0x44,0x45,0x45,0x46,0x46,0x47,0x47,
										0x48,0x48,0x49,0x49,0x4a,0x4a,0x4b,0x4b,0x4c,0x4c,0x4d,0x4d,0x4e,0x4e,0x4f,0x4f,
										0x50,0x50,0x51,0x51,0x52,0x52,0x53,0x53,0x54,0x54,0x55,0x55,0x56,0x56,0x57,0x57,
										0x58,0x58,0x59,0x59,0x5a,0x5a,0x5b,0x5b,0x5c,0x5c,0x5d,0x5d,0x5e,0x5e,0x5f,0x5f,
										0x60,0x60,0x61,0x61,0x62,0x62,0x63,0x63,0x64,0x64,0x65,0x65,0x66,0x66,0x67,0x67,
										0x68,0x68,0x69,0x69,0x6a,0x6a,0x6b,0x6b,0x6c,0x6c,0x6d,0x6d,0x6e,0x6e,0x6f,0x6f,
										0x70,0x70,0x71,0x71,0x72,0x72,0x73,0x73,0x74,0x74,0x75,0x75,0x76,0x76,0x77,0x77,
										0x78,0x78,0x79,0x79,0x7a,0x7a,0x7b,0x7b,0x7c,0x7c,0x7d,0x7d,0x7e,0x7e,0x7f,0x7f,

										// RBORDER
										0x80,0x80,0x81,0x81,0x82,0x82,0x83,0x83,0x84,0x84,0x85,0x85,0x86,0x86,0x87
									} ;

//! V カウンタ
static uint8_t						m_vcount[VDP_VSCREEN_HEIGHT]={
										// BBLANKING
										0xd8,0xd9,0xda,

										// VSYNC
										0xd5,0xd6,0xd7,

										// TBLANKING
										0xd8,0xd9,0xda,0xdb,0xdc,0xdd,0xde,0xdf,0xe0,0xe1,0xe2,0xe3,0xe4,

										// TBORDER
										0xe5,0xe6,0xe7,0xe8,0xe9,0xea,0xeb,0xec,0xed,0xee,0xef,0xf0,0xf1,0xf2,0xf3,
										0xf4,0xf5,0xf6,0xf7,0xf8,0xf9,0xfa,0xfb,0xfc,0xfd,0xfe,0xff,

										// VACTIVE
										0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
										0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,
										0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,
										0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f,
										0x40,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4a,0x4b,0x4c,0x4d,0x4e,0x4f,
										0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x5b,0x5c,0x5d,0x5e,0x5f,
										0x60,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,
										0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x7b,0x7c,0x7d,0x7e,0x7f,
										0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
										0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0x9f,
										0xa0,0xa1,0xa2,0xa3,0xa4,0xa5,0xa6,0xa7,0xa8,0xa9,0xaa,0xab,0xac,0xad,0xae,0xaf,
										0xb0,0xb1,0xb2,0xb3,0xb4,0xb5,0xb6,0xb7,0xb8,0xb9,0xba,0xbb,0xbc,0xbd,0xbe,0xbf,

										// BBORDER
										0xc0,0xc1,0xc2,0xc3,0xc4,0xc5,0xc6,0xc7,0xc8,0xc9,0xca,0xcb,0xcc,0xcd,0xce,0xcf,
										0xd0,0xd1,0xd2,0xd3,0xd4,0xd5,0xd6,0xd7
									} ;

//! 行情報
static uint16_t						m_lineInfo[]={
										// BOTTOM BLANKING + VERTICAL SYNC + TOP BLANKING
										LINEDATA(RBLANK_BEGIN, 47), LINEDATA(VCOUNT, 295),																	//   0
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   1
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   2
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   3
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   4
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   5
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   6
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   7
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   8
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//   9
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  10
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  11
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  12
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  13
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  14
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  15
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  16
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  17
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  18

										// TOP BORDER
										LINEDATA(RBLANK,       47), LINEDATA(VCOUNT, 295),																	//  19
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  20
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  21
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  22
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  23
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  24
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  25
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  26
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  27
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  28
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  29
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  30
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  31
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  32
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  33
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  34
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  35
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  36
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  37
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  38
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  39
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  40
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  41
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  42
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  43
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(VCOUNT,   2), LINEDATA(OVR_FLAG, 293),											//  44
										LINEDATA(RBLANK_DRAW,  47), LINEDATA(YPOS,     2), LINEDATA(OVR_FLAG, 293),											//  45

										// ACTIVE BORDER
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  46
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  47
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  48
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  49
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  50
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  51
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  52
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  53
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  54
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  55
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  56
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  57
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  58
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  59
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  60
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  61
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  62
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  63
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  64
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  65
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  66
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  67
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  68
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  69
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  70
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  71
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  72
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  73
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  74
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  75
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  76
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  77
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  78
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  79
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  80
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  81
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  82
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  83
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  84
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  85
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  86
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  87
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  88
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  89
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  90
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  91
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  92
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  93
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  94
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  95
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  96
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  97
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  98
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						//  99
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 100
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 101
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 102
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 103
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 104
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 105
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 106
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 107
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 108
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 109
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 110
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 111
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 112
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 113
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 114
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 115
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 116
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 117
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 118
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 119
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 120
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 121
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 122
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 123
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 124
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 125
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 126
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 127
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 128
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 129
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 130
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 131
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 132
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 133
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 134
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 135
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 136
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 137
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 138
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 139
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 140
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 141
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 142
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 143
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 144
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 145
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 146
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 147
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 148
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 149
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 150
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 151
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 152
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 153
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 154
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 155
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 156
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 157
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 158
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 159
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 160
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 161
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 162
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 163
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 164
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 165
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 166
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 167
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 168
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 169
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 170
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 171
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 172
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 173
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 174
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 175
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 176
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 177
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 178
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 179
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 180
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 181
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 182
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 183
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 184
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 185
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 186
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 187
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 188
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 189
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 190
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 191
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 192
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 193
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 194
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 195
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 196
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 197
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 198
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 199
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 200
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 201
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 202
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 203
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 204
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 205
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 206
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 207
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 208
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 209
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 210
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 211
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 212
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 213
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 214
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 215
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 216
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 217
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 218
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 219
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 220
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 221
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 222
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 223
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 224
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 225
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 226
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 227
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 228
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 229
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 230
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 231
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 232
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 233
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 234
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 235
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 236
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 2), LINEDATA(OVR_FLAG, 293),						// 237

										// BOTTOM BORDER
										LINEDATA(RBLANK_DRAW, 46), LINEDATA(HSYNC, 1), LINEDATA(VCOUNT, 295),												// 238
										LINEDATA(RBLANK_DRAW, 45), LINEDATA(VSYNC, 2), LINEDATA(VCOUNT,   2), LINEDATA(VSYNC_FLAG, 293),					// 239
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 240
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 241
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 242
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 243
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 244
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 245
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 246
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 247
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 248
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 249
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 250
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 251
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 252
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 253
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 254
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 255
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 256
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 257
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 258
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 259
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 260
										LINEDATA(RBLANK_DRAW, 47),                     LINEDATA(VCOUNT, 295),												// 261

										LINEDATA(RBLANK_DRAW_END, -47),
									} ;


// -------------------------------------------------------------------------------- 
//!  16 バイト単位のメモリー設定                                                    
//!                                                                                 
//!  @param [out] p_dest  設定先アドレス                                            
//!  @param [in]  value   設定値                                                    
//!  @param [in]  size    クリアサイズ                                              
//!  @return              設定後の設定先アドレス                                    
// -------------------------------------------------------------------------------- 
static inline uint32_t *memset16(uint32_t *p_dest, uint32_t value, uint size)
{
	uint32_t	*p_destEnd ;


	p_destEnd = (uint32_t *)&((uint8_t *)p_dest)[size] ;

	do {
		p_dest[0] = value ;
		p_dest[1] = value ;
		p_dest[2] = value ;
		p_dest[3] = value ;
		p_dest   += 4 ;
	} while( p_dest != p_destEnd ) ;

	return p_dest ;
}


// -------------------------------------------------------------------------------- 
//!  32 バイト単位のメモリーコピー                                                  
//!                                                                                 
//!  @param [out] p_dest  コピー先アドレス                                          
//!  @param [in]  p_srce  コピー元アドレス                                          
//!  @param [in]  size    コピーサイズ                                              
// -------------------------------------------------------------------------------- 
static inline void memcpy32(uint32_t *p_dest, uint32_t const *p_srce, uint size)
{
	uint32_t	*p_destEnd ;


	p_destEnd = (uint32_t *)&((uint8_t *)p_dest)[size] ;

	do {
		p_dest[0] = p_srce[0] ;
		p_dest[1] = p_srce[1] ;
		p_dest[2] = p_srce[2] ;
		p_dest[3] = p_srce[3] ;
		p_dest[4] = p_srce[4] ;
		p_dest[5] = p_srce[5] ;
		p_dest[6] = p_srce[6] ;
		p_dest[7] = p_srce[7] ;
		p_srce   += 8 ;
		p_dest   += 8 ;
	} while( p_dest != p_destEnd ) ;
}


// -------------------------------------------------------------------------------- 
//!  スプライト情報作成                                                             
//!                                                                                 
//!  @param [in,out] p             this ポインタ                                    
//!  @param [out]    p_spriteInfo  スプライト情報格納アドレス                       
//!  @param [in]     line          描画行                                           
//!  @return                       なし                                             
// -------------------------------------------------------------------------------- 
static inline void RAMFUNC createSpriteInfo(VDP *p, SPRITEINFO *p_spriteInfo, uint line)
{
	uint8_t		*p_spriteAttr, *p_plane, *p_pattern ;
	uint32_t	pixel, name ;
	int			left, xofs, xpos, ypos, shift ;
	uint		index, count, height ;


	count        = 0 ;
	p_spriteAttr = &p->vram[(uint )(p->regs[5]&0x7e)<< 7] ;
	p_pattern    = &p->vram[(uint )(p->regs[6]&0x04)<<11] ;
	left         = (p->regs[0] & R0_LCB) >> 2 ;
	xofs         = (p->regs[0] & R0_SS) + left ;

	switch( p->regs[1] & (R1_SIZE|R1_MAG) ) {
		case 0:			// 通常 8x8
		case R1_SIZE:	// 通常 8x16
		{
			height = 8 ;
			if( p->regs[1] & R1_SIZE ) {
				height = 16 ;
			}

			for(index=0; index<VDP_SPRITE_MAX; index++) {
				ypos = p_spriteAttr[index] ;
				if( ypos == 208 ) {
					break ;
				}

				// ｙ座標
				ypos = (line + height) - (uint8_t )(ypos + height + 1) ;
				if( (uint )ypos >= height ) {
					continue ;
				}

				xpos = p_spriteAttr[0x80+(index*2)] - xofs ;
				if( (xpos+7) < 0 ) {
					continue ;
				}

				// 表示数オーバー
				if( count == 8 ) {

					// オーバーフラグは割り込みフラグがクリアされている時のみ
					if( p->status & STATUS_F) {

						// 非表示位置はオーバーフラグを無視（衝突フラグは立つ）
						if( (uint8_t )(p_spriteAttr[index]-191) > (uint8_t )(255-191-height) ) {
							p->status |= STATUS_5S ;
						}
					}
					break ;
				}

				name = p_spriteAttr[0x81+(index*2)] ;
				if( p->regs[1] & R1_SIZE ) {
					name &= ~1 ;
				}

				p_plane = &p_pattern[((uint )name*32)|((ypos&(height-1))*4)] ;
				pixel   = (m_planeToPixel[0][p_plane[0]] << 0) |
				          (m_planeToPixel[0][p_plane[1]] << 1) |
				          (m_planeToPixel[0][p_plane[2]] << 2) |
				          (m_planeToPixel[0][p_plane[3]] << 3) ;

				if( xpos < 0 ) {
					pixel <<= (-xpos * 4) ;
					xpos    = 0 ;
				}

				xpos += left ;

				if( xpos > 248 ) {
					pixel &= (0xffffffff << ((xpos-248)*4)) ;
				}

				p_spriteInfo->xpos       = xpos ;
				p_spriteInfo->pattern[0] = pixel ;
				p_spriteInfo->pattern[1] = 0 ;
				p_spriteInfo++ ;
				count++ ;
			}
			break ;
		}

		default:		// 拡大 8x16
		{
			height = 16 ;
			if( p->regs[1] & R1_SIZE ) {
				height = 32 ;
			}

			for(index=0; index<VDP_SPRITE_MAX; index++) {
				ypos = p_spriteAttr[index] ;
				if( ypos == 208 ) {
					break ;
				}

				// ｙ座標
				ypos = (line + height) - (uint8_t )(ypos + height + 1) ;
				if( (uint )ypos >= height ) {
					continue ;
				}
				ypos /= 2 ;

				xpos = p_spriteAttr[0x80+(index*2)] - xofs ;
				if( (xpos+16) < 0 ) {
					continue ;
				}

				// 表示数オーバー
				if( count == 8 ) {

					// 非表示位置はオーバーフラグを無視（衝突フラグは立つ）
					if( (uint8_t )(p_spriteAttr[index]-191) > (uint8_t )(255-191-height) ) {
						p->status |= STATUS_5S ;
					}
					break ;
				}

				name = p_spriteAttr[0x81+(index*2)] ;
				if( p->regs[1] & R1_SIZE ) {
					name &= ~1 ;
				}

				p_plane = &p_pattern[((uint )name*32)|((ypos&(height-1))*4)] ;
				pixel   = (m_planeToPixel[0][p_plane[0]] << 0) |
				          (m_planeToPixel[0][p_plane[1]] << 1) |
				          (m_planeToPixel[0][p_plane[2]] << 2) |
				          (m_planeToPixel[0][p_plane[3]] << 3) ;

				p_spriteInfo->pattern[0]  = ((pixel & 0xf0000000)     >>  4) |
				                            ((pixel & 0x0f000000)     >>  8) |
				                            ((pixel & 0x00f00000)     >> 12) |
				                            ((pixel & 0x000f0000)     >> 16) ;
				p_spriteInfo->pattern[0] |= (p_spriteInfo->pattern[0] <<  4) ;

				p_spriteInfo->pattern[1]  = ((pixel & 0x0000f000)     << 12) |
				                            ((pixel & 0x00000f00)     <<  8) |
				                            ((pixel & 0x000000f0)     <<  4) |
				                            ((pixel & 0x0000000f)     <<  0) ;
				p_spriteInfo->pattern[1] |= (p_spriteInfo->pattern[1] <<  4) ;

				if( xpos < 0 ) {
					shift                    = -xpos * 4 ;
					p_spriteInfo->pattern[0] = (p_spriteInfo->pattern[0] << shift) | (p_spriteInfo->pattern[1] >> (32-shift)) ;
					p_spriteInfo->pattern[1] = (p_spriteInfo->pattern[1] << shift) ;
					xpos                     = 0 ;
				}

				xpos += left ;

				if( xpos > 240 ) {
					if( xpos < 248 ) {
						p_spriteInfo->pattern[1] &= (0xffffffff << ((xpos-248)*4)) ;
					} else {
						p_spriteInfo->pattern[0] &= (0xffffffff << ((xpos-240)*4)) ;
						p_spriteInfo->pattern[1]  = 0 ;
					}
				}

				p_spriteInfo->xpos = xpos ;
				p_spriteInfo++ ;
				count++ ;
			}
			break ;
		}
	}

	p->spriteInfoCount = count ;
}


// -------------------------------------------------------------------------------- 
//!  スプライト状態を更新                                                           
//!                                                                                 
//!  @param [in,out] p  this ポインタ                                               
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
static inline void RAMFUNC updateSpriteStatus( VDP *p )
{
	SPRITEINFO	*p_spriteInfo, *p_spriteInfoEnd ;
	uint32_t	drawFlag[SM3_WIDTH/32] ;
	uint32_t	pattern ;
	int			xpos, xhit ;


	if( !p->spriteInfoCount ) {
		return ;
	}

	// 非表示時はスプライトオーバーは発生しない？
	if( !p->blank ) {
		return ;
	}

	if( p->status & STATUS_C ) {
		return ;
	}

	p_spriteInfo    = &p->spriteInfo[0] ;
	p_spriteInfoEnd = &p->spriteInfo[p->spriteInfoCount] ;
	drawFlag[0]     = 0 ;
	drawFlag[1]     = 0 ;
	drawFlag[2]     = 0 ;
	drawFlag[3]     = 0 ;
	drawFlag[4]     = 0 ;
	drawFlag[5]     = 0 ;
	drawFlag[6]     = 0 ;
	drawFlag[7]     = 0 ;
	xhit            = SM3_WIDTH ;

	do {
		pattern = p_spriteInfo->pattern[0] ;
		if( pattern ) {
			xpos = p_spriteInfo->xpos ;

			while( pattern ) {
				if( pattern >> 28 ) {
					if( drawFlag[xpos/32] & (1<<(xpos&31)) ) {
						if( xhit > xpos ) {
							xhit = xpos ;
							break ;
						}
					}
					drawFlag[xpos/32] |= (1<<(xpos&31)) ;
				}
				pattern <<= 4 ;
				xpos++ ;
			}
		}
	} while( ++p_spriteInfo != p_spriteInfoEnd ) ;

	if( xhit != SM3_WIDTH ) {
		clockgen_attachEvent(&p->clockEventOverlap, 34+xhit) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  スプライトを描画                                                               
//!                                                                                 
//!  @param [in,out] p      this ポインタ                                           
//!  @param [in]     p_buff 描画先バッファ                                          
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static inline void RAMFUNC drawSprite(VDP *p, uint8_t *p_buff)
{
	SPRITEINFO	*p_spriteInfo ;
	uint8_t		*p_dest ;
	uint32_t	value, pattern ;
	int			xpos ;
	uint		index ;

	// 初期化
	p_buff      += 8 ;
	p_spriteInfo = &p->spriteInfo[0] ;

	// 描画
	for(index=0; index<p->spriteInfoCount; index++,p_spriteInfo++) {
		// 左側
		pattern = p_spriteInfo->pattern[0] ;
		if( pattern ) {
			xpos   = p_spriteInfo->xpos ;
			p_dest = &p_buff[xpos] ;

			while( pattern ) {
				value = pattern >> 28 ;
				if( value ) {
					if( (*p_dest&0x4f) <= 0x40 ) {
						*p_dest = (uint8_t )(value | (0x30+0x20)) ;
					}
				}
				pattern <<= 4 ;
				xpos++ ;
				p_dest++ ;
			}
		}

		// 右側
		pattern = p_spriteInfo->pattern[1] ;
		if( pattern ) {
			xpos   = p_spriteInfo->xpos + 8 ;
			p_dest = &p_buff[xpos] ;

			while( pattern ) {
				value = pattern >> 28 ;
				if( value ) {
					if( (*p_dest&0x4f) <= 0x40 ) {
						*p_dest = (uint8_t )(value | (0x30+0x20)) ;
					}
				}
				pattern <<= 4 ;
				xpos++ ;
				p_dest++ ;
			}
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  作業バッファの内容をコピー                                                     
//!                                                                                 
//!  @param [out] p_dest  コピー先アドレス                                          
//!  @param [in]  p_buff  作業バッファアドレス                                      
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void RAMFUNC copyBuff(uint8_t *p_dest, uint8_t const *p_buff)
{
	uint8_t		*p_destEnd ;


	if( g.mode == MODE_SM3 ) {
		p_destEnd = &p_dest[SM3_WIDTH] ;
	} else {
		// 左右非表示
		memset16((uint32_t *)&p_dest[0],      0x00000000, GG_MARGIN_LEFT) ;
		memset16((uint32_t *)&p_dest[64+160], 0x00000000, GG_MARGIN_RIGHT) ;

		p_dest   += GG_MARGIN_LEFT ;
		p_destEnd = &p_dest[GG_WIDTH] ;
		p_buff   += GG_MARGIN_LEFT ;
	}

	do {
		*(uint32_t *)&p_dest[0]  = *(uint32_t *)&p_buff[0]  & 0x3f3f3f3f ;
		*(uint32_t *)&p_dest[4]  = *(uint32_t *)&p_buff[4]  & 0x3f3f3f3f ;
		*(uint32_t *)&p_dest[8]  = *(uint32_t *)&p_buff[8]  & 0x3f3f3f3f ;
		*(uint32_t *)&p_dest[12] = *(uint32_t *)&p_buff[12] & 0x3f3f3f3f ;
		*(uint32_t *)&p_dest[16] = *(uint32_t *)&p_buff[16] & 0x3f3f3f3f ;
		*(uint32_t *)&p_dest[20] = *(uint32_t *)&p_buff[20] & 0x3f3f3f3f ;
		*(uint32_t *)&p_dest[24] = *(uint32_t *)&p_buff[24] & 0x3f3f3f3f ;
		*(uint32_t *)&p_dest[28] = *(uint32_t *)&p_buff[28] & 0x3f3f3f3f ;
		p_buff                   += 32 ;
		p_dest                   += 32 ;
	} while( p_dest != p_destEnd ) ;
}


// -------------------------------------------------------------------------------- 
//!  １行描画                                                                       
//!                                                                                 
//!  @param [in,out] p     this ポインタ                                            
//!  @param [in]     line  描画行                                                   
// -------------------------------------------------------------------------------- 
static inline void RAMFUNC drawLine(VDP *p, uint line)
{
	VRAM		*p_vram ;
	uint8_t		xpos, *p_srce, *p_nt[2], *p_buff ;
	uint16_t	pattern ;
	uint32_t	hi, pixel, *p_planeToPixel, buff[(8+SM3_WIDTH+8)/4], *p_dest32 ;
	uint		index, count, ypos, pos ;

	// 上下ボーダー
	line -= VDP_TBORDER ;
	if( line >= VDP_VACTIVE ) {
		return ;
	}

	p_vram = &g.vram[g.page] ;

	// パレットコピー
	memcpy32((uint32_t *)&p_vram->palette[line][0], (uint32_t const *)&p->palette[0], sizeof( p_vram->palette[0] )) ;

	// GAME GEAR 時は上下非表示
	if( g.mode == MODE_GG ) {
		if( (uint )(line-GG_MARGIN_TOP) >= GG_HEIGHT ) {
			memset16((uint32_t *)&p_vram->image[line][0], 0x00000000, SM3_WIDTH) ;
			return ;
		}
	}

	// 非表示
	if( !p->blank ) {
		if( g.mode == MODE_SM3 ) {
			memset16((uint32_t *)&p_vram->image[line][0], p->borderColor, SM3_WIDTH) ;
		} else {
			p_dest32 = (uint32_t *)&p_vram->image[line][0] ;
			p_dest32 = memset16(&p_dest32[0], 0x00000000,     GG_MARGIN_LEFT) ;
			p_dest32 = memset16(&p_dest32[0], p->borderColor, GG_WIDTH) ;
			p_dest32 = memset16(&p_dest32[0], 0x00000000,     GG_MARGIN_RIGHT) ;
		}
		return ;
	}

	p_buff = (uint8_t *)&buff[8/4] ;
	xpos   = (uint8_t )-p->sx ;
	ypos   = (uint8_t )p->sy ;
	index  = 0 ;
	count  = (SM3_WIDTH / 8) + ((xpos & 7) != 0) ;

	// 上側固定
	if( (line<16) && (p->regs[0]&R0_HSI) ) {
		xpos = 0 ;
	}

	ypos += line ;
	if( ypos >= 224 ) {
		ypos -= 224 ;
	}

	p_buff -= (xpos & 7) ;

	// 左ボーダー無視
	if( p->regs[0] & R0_LCB ) {
		p_buff += 8 ;
		xpos   += 8 ;
		index   = 1 ;
	}

	if( p->nt & (1<<0) ) {
		p_nt[0] = &p->vram[(((uint16_t )(p->nt&0x0e)<<10)+((line/8)*64))] ;
		p_nt[1] = &p->vram[(((uint16_t )(p->nt&0x0e)<<10)+((ypos/8)*64))] ;
	} else {
		p_nt[0] = &p->vram[(((uint16_t )(p->nt&0x0e)<<10)+((line/8)*64))&~0x0400] ;
		p_nt[1] = &p->vram[(((uint16_t )(p->nt&0x0e)<<10)+((ypos/8)*64))&~0x0400] ;
	}

	// 背景描画
	for(; index<count; index++,xpos+=8) {
		if( (index>=(SM3_HEIGHT/8)) && (p->regs[0]&R0_VSI) ) {
			pattern = ((uint16_t )p_nt[0][(xpos/8)*2+1] << 8) | p_nt[0][(xpos/8)*2+0] ;
			pos     = line & 7 ;
		} else {
			pattern = ((uint16_t )p_nt[1][(xpos/8)*2+1] << 8) | p_nt[1][(xpos/8)*2+0] ;
			pos     = ypos & 7 ;
		}

		if( !(pattern&(1<<10)) ) {		// 通常
			p_srce = &p->vram[((pattern&511)*32)+(   pos *4)] ;
		} else {						// 垂直反転
			p_srce = &p->vram[((pattern&511)*32)+((7-pos)*4)] ;
		}

		p_planeToPixel = &m_planeToPixel[(pattern>>9)&1][0] ;
		pixel          = (p_planeToPixel[p_srce[0]] << 0) |
						 (p_planeToPixel[p_srce[1]] << 1) |
						 (p_planeToPixel[p_srce[2]] << 2) |
						 (p_planeToPixel[p_srce[3]] << 3) ;

		hi = ((pattern >> 6) & 0x40) | ((pattern >> 7) & 0x10) | 0x20 ;

		p_buff[0] = (uint8_t )(( pixel        >> 28) | hi) ;
		p_buff[1] = (uint8_t )(((pixel <<  4) >> 28) | hi) ;
		p_buff[2] = (uint8_t )(((pixel <<  8) >> 28) | hi) ;
		p_buff[3] = (uint8_t )(((pixel << 12) >> 28) | hi) ;
		p_buff[4] = (uint8_t )(((pixel << 16) >> 28) | hi) ;
		p_buff[5] = (uint8_t )(((pixel << 20) >> 28) | hi) ;
		p_buff[6] = (uint8_t )(((pixel << 24) >> 28) | hi) ;
		p_buff[7] = (uint8_t )(((pixel << 28) >> 28) | hi) ;
		p_buff   += 8 ;
	}

	// スプライト描画
	if( p->spriteInfoCount ) {
		drawSprite(p, (uint8_t *)&buff[0]) ;
	}

	// 左ボーダー
	if( p->regs[0] & R0_LCB ) {
		buff[8/4+0] = p->borderColor ;
		buff[8/4+1] = p->borderColor ;
	}

	// 作業バッファから VRAM へコピー
	copyBuff(&p_vram->image[line][0], (uint8_t *)&buff[8/4]) ;
}


static int32_t RAMFUNC clockevent_process_rblank_begin( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_rblank( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_rblank_draw( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_rblank_draw_end( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_vcount( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_vsync( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_vsync_flag( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_hsync( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_ypos( CLOCKEVENT *p_this ) ;
static int32_t RAMFUNC clockevent_process_ovr_flag( CLOCKEVENT *p_this ) ;


static CLOCKEVENTPROC				m_lineProc[]={					//!< 行処理
										clockevent_process_rblank_begin,
										clockevent_process_rblank,
										clockevent_process_rblank_draw,
										clockevent_process_rblank_draw_end,
										clockevent_process_vcount,
										clockevent_process_vsync,
										clockevent_process_vsync_flag,
										clockevent_process_hsync,
										clockevent_process_ypos,
										clockevent_process_ovr_flag
									} ;


// -------------------------------------------------------------------------------- 
//!  次のイベントを設定                                                             
//!                                                                                 
//!  @param [in,out] p  this ポインタ                                               
//!  @return            クロック数                                                  
// -------------------------------------------------------------------------------- 
static inline int32_t RAMFUNC setNextEvent( VDP *p )
{
	int32_t		clock ;


	clock                 = ((int32_t )(int16_t )*p->p_lineInfo++>>4) ;
	p->clockEvent.process = m_lineProc[*p->p_lineInfo&15] ;

	return clock ;
}


// -------------------------------------------------------------------------------- 
//!  クロックイベント処理                                                           
//!                                                                                 
//!  @param [in,out] p_this    this ポインタ                                        
//!  @retval         0         クロックイベント破棄                                 
//!  @retval         > 0       クロック数                                           
//!  @retval         < 0       クロック数（符号反転）＋クロックジェネレータ停止     
// -------------------------------------------------------------------------------- 
static int32_t RAMFUNC clockevent_process_overlap( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEventOverlap, p_this) ;


	p->status |= STATUS_C ;

	return 0 ;
}

static int32_t RAMFUNC clockevent_process_rblank_begin( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEvent, p_this) ;


	p->beginDrawClock = g.baseClock ;
	p->scanLine       = 0 ;

	return setNextEvent( p ) ;
}

static int32_t RAMFUNC clockevent_process_rblank( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEvent, p_this) ;


	p->beginDrawClock = g.baseClock ;
	p->scanLine++ ;

	return setNextEvent( p ) ;
}

static int32_t RAMFUNC clockevent_process_rblank_draw( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEvent, p_this) ;


	if( p->regs[0] & R0_M4 ) {
		drawLine(p, p->scanLine-VDP_TMARGIN) ;
	} else {
//#		vdmod_tms9918_core_drawLine(&p->core, p->scanLine-VDP_TMARGIN, &p->image[p->scanLine-VDP_TMARGIN][0]) ;
	}

	p->beginDrawClock = g.baseClock ;
	p->scanLine++ ;

	return setNextEvent( p ) ;
}

static int32_t RAMFUNC clockevent_process_rblank_draw_end( CLOCKEVENT *p_this )
{
	VDP			*p=_THISPTR(VDP, clockEvent, p_this) ;
	int32_t		clock ;


	if( p->regs[0] & R0_M4 ) {
		drawLine(p, p->scanLine-VDP_TMARGIN) ;
	} else {
//#		vdmod_tms9918_core_drawLine(&p->core, p->scanLine-VDP_TMARGIN, &p->image[p->scanLine-VDP_TMARGIN][0]) ;
	}

	p->beginDrawClock = g.baseClock ;
	p->scanLine       = 0 ;

	clock                 = ((int32_t )(int16_t )*p->p_lineInfo>>4) ;
	p->p_lineInfo         = &m_lineInfo[1] ;
	p->clockEvent.process = m_lineProc[*p->p_lineInfo&15] ;

	return clock ;
}

static int32_t RAMFUNC clockevent_process_vcount( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEvent, p_this) ;


	p->vcount = m_vcount[p->scanLine] ;

	return setNextEvent( p ) ;
}

static int32_t RAMFUNC clockevent_process_vsync( CLOCKEVENT *p_this )
{
	VDP			*p=_THISPTR(VDP, clockEvent, p_this) ;
	uint32_t	borderColor ;

	// VSync 割り込み
	if( p->regs[1] & R1_IE ) {
		writePin(p->p_pin, 0) ;
	}

	g.vram[g.page].borderColor = p->palette[16+(p->regs[7]&15)] ;
	borderColor                = (0x20 + 16) + (p->regs[7] & 15) ;
	borderColor               |= (borderColor <<  8) ;
	borderColor               |= (borderColor << 16) ;
	p->borderColor             = borderColor ;

	p->blank           = p->regs[1] & R1_BLANK ;
	p->spriteInfoCount = 0 ;

	return setNextEvent( p ) ;
}

static int32_t RAMFUNC clockevent_process_vsync_flag( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEvent, p_this) ;


	p->status |= STATUS_F ;

	return setNextEvent( p ) ;
}

static int32_t RAMFUNC clockevent_process_hsync( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEvent, p_this) ;

	// HSync 割り込み
	if( !p->lineCount-- ) {
		p->lineCount = p->regs[10] ;
		if( p->regs[0] & R0_IE1 ) {
			writePin(p->p_pin, 0) ;
		}
	}

	p->blank = p->regs[1] & R1_BLANK ;
	p->nt    = p->regs[2] ;
	p->sx    = p->regs[8] ;

	return setNextEvent( p ) ;
}

static int32_t RAMFUNC clockevent_process_ypos( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEvent, p_this) ;


	p->vcount = m_vcount[p->scanLine] ;

	p->sy         = p->regs[9] ;
	p->lineCount  = p->regs[10] ;
	p->blank      = p->regs[1] & R1_BLANK ;

	return setNextEvent( p ) ;
}

static int32_t RAMFUNC clockevent_process_ovr_flag( CLOCKEVENT *p_this )
{
	VDP		*p=_THISPTR(VDP, clockEvent, p_this) ;


	createSpriteInfo(p, &p->spriteInfo[0], (uint8_t )(p->scanLine-(VDP_TMARGIN+VDP_TBORDER))) ;

	updateSpriteStatus( p ) ;

	return setNextEvent( p ) ;
}


// -------------------------------------------------------------------------------- 
//!  アドレスバス読み込み                                                           
//!                                                                                 
//!  @param [in,out] p_this   this ポインタ                                         
//!  @param          address  アドレス                                              
//!  @return         データ                                                         
// -------------------------------------------------------------------------------- 
static uint8_t RAMFUNC addressbus_read(ADDRESSBUS *p_this, uint16_t address)
{
	VDP			*p=_THISPTR(VDP, addressBus, p_this) ;
	uint8_t		data ;


	switch( address & 3 ) {
		case 0:		// データ読み込み
		{
			p->write = 0 ;

			data        = p->readData ;
			p->readData = p->vram[p->address] ;
			p->address  = (p->address + 1) & 0x3fff ;
			break ;
		}

		case 1:		// ステータス読み込み
		{
			p->write = 0 ;

			data      = p->status ;
			p->status = 0 ;

			writePin(p->p_pin, 1) ;
			break ;
		}

		case 2:		// V Counter
		{
			data = p->vcount ;
			break ;
		}

		default:	// H Counter
		{
			data = m_hcount[g.baseClock-p->beginDrawClock] ;
			break ;
		}
	}

	return data ;
}


// -------------------------------------------------------------------------------- 
//!  アドレスバス書き込み                                                           
//!                                                                                 
//!  @param [in,out] p_this   this ポインタ                                         
//!  @param          address  アドレス                                              
//!  @param          データ                                                         
//!  @return         なし                                                           
// -------------------------------------------------------------------------------- 
static void RAMFUNC addressbus_write(ADDRESSBUS *p_this, uint16_t address, uint8_t data)
{
	VDP		*p=_THISPTR(VDP, addressBus, p_this) ;
	uint	index ;


	switch( address & 3 ) {
		case 0:		// データ書き込み
		{
			p->write    = 0 ;
			p->readData = data ;

			if( p->cmd != CMD_WRITECRAM ) {		// VRAM 書き込み
				p->vram[p->address] = data ;
			} else {							// CRAM 書き込み
				if( g.mode == MODE_SM3 ) {		// SEGA MARK III
					p->palette[p->address&31] = (((uint16_t )(data & 0x03) << 0) * 5) |		// R
					                            (((uint16_t )(data & 0x0c) << 2) * 5) |		// G
					                            (((uint16_t )(data & 0x30) << 4) * 5) ;		// B
				} else {						// GAME GEAR
					if( !(p->address&1) ) {
						p->colorGG = data ;
					} else {
						p->palette[(p->address/2)&31] = ((uint16_t )(data & 0x0f) << 8) | p->colorGG ;
					}
				}
			}
			p->address = (p->address + 1) & 0x3fff ;
			break ;
		}

		case 1:		// コマンド書き込み
		{
			p->write ^= 1 ;

			if( p->write ) {
				p->address   = (p->address & 0x3f00) | data ;
				p->writeData = data ;
				break ;
			}

			p->cmd = data >> 6 ;

			p->address = ((uint16_t )(data & 0x3f) << 8) | (uint8_t )p->address ;

			switch( p->cmd ) {
				case CMD_READVRAM:
				{
					p->readData = p->vram[p->address] ;
					p->address  = (p->address + 1) & 0x3fff ;
					break ;
				}

				case CMD_WRITEREG:
				{
					index = data & 0x0f ;

					// 画面制御レジスタ１
					if( !index ) {
						if( !(p->status&STATUS_F) ) {
							if( ((p->regs[0]^p->writeData)&p->regs[0]) & R0_IE1 ) {
								p->p_pin->write(p->p_pin, 1) ;
							}
						}
					}

					p->regs[index] = p->writeData ;
					break ;
				}

//				case CMD_WRITECRAM:
				default:
					break ;
			}
			break ;
		}

		default:
			break ;
	}
}


// -------------------------------------------------------------------------------- 
//!   初期化                                                                        
//!                                                                                 
//!   @param [in,out] p       this ポインタ                                         
//!   @param [in]     p_pint  割り込みピン                                          
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
void vdp_init(VDP *p, PIN *p_pin)
{
	int		index ;


	p->p_pin            = p_pin ;
	p->addressBus.read  = addressbus_read ;
	p->addressBus.write = addressbus_write ;

	for(index=0; index<256; index++) {

		// 通常 
		m_planeToPixel[0][index] = ((uint32_t )((index >> 7) & 1) <<28) |
		                           ((uint32_t )((index >> 6) & 1) <<24) |
		                           ((uint32_t )((index >> 5) & 1) <<20) |
		                           ((uint32_t )((index >> 4) & 1) <<16) |
		                           ((uint32_t )((index >> 3) & 1) <<12) |
		                           ((uint32_t )((index >> 2) & 1) << 8) |
		                           ((uint32_t )((index >> 1) & 1) << 4) |
		                           ((uint32_t )((index >> 0) & 1) << 0) ;

		// 左右反転 
		m_planeToPixel[1][index] = ((uint32_t )((index >> 7) & 1) << 0) |
		                           ((uint32_t )((index >> 6) & 1) << 4) |
		                           ((uint32_t )((index >> 5) & 1) << 8) |
		                           ((uint32_t )((index >> 4) & 1) <<12) |
		                           ((uint32_t )((index >> 3) & 1) <<16) |
		                           ((uint32_t )((index >> 2) & 1) <<20) |
		                           ((uint32_t )((index >> 1) & 1) <<24) |
		                           ((uint32_t )((index >> 0) & 1) <<28) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  リセット                                                                       
//!                                                                                 
//!  @param [in,out] p  this ポインタ                                               
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void vdp_reset( VDP *p )
{
	p->status                    = 0 ;
	p->write                     = 0 ;
	p->writeData                 = 0 ;
	p->readData                  = 0 ;
	p->cmd                       = 0 ;
	p->address                   = 0 ;
	p->regs[0]                   = 0x36 ;
	p->regs[1]                   = 0x80 ;
	p->regs[2]                   = 0xff ;
	p->regs[3]                   = 0xff ;
	p->regs[4]                   = 0xff ;
	p->regs[5]                   = 0xff ;
	p->regs[6]                   = 0xfb ;
	p->regs[7]                   = 0x00 ;
	p->regs[8]                   = 0x00 ;
	p->regs[9]                   = 0x00 ;
	p->regs[10]                  = 0xff ;
	p->p_lineInfo                = &m_lineInfo[1] ;
	p->clockEvent.process        = m_lineProc[m_lineInfo[0]&15] ;
	p->clockEventOverlap.process = clockevent_process_overlap ;

	writePin(p->p_pin, 1) ;

	clockgen_attachEvent(&p->clockEvent, 256) ;
}
